/* xkiss 2.00 test c (KISS GS4) 1993/2/28 */

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <ctype.h>


typedef enum {OFF,ON} STATE;

/* if you want, set default status ON or OFF */
STATE	big_mode     = OFF;	/* -b option */
STATE	quiet_mode   = OFF;	/* -q option */
STATE	iconic_mode  = OFF;	/* -i option */
STATE	reverse_mode = OFF;	/* -r option */

/* if you want, set default color mode. ON state is only one */
STATE	color_mode       = ON;	/* -c option */
STATE	mono_dither_mode = OFF;	/* -m option */
STATE	mono_gosa_mode   = OFF;	/* -g option */


/* kiss GS definition */

#define	GS1_SIZE_X	640
#define	GS2_SIZE_X	640
#define	GS3_SIZE_X	640
#define	GS4_SIZE_X	768

#define	GS1_SIZE_Y	400
#define	GS2_SIZE_Y	400
#define	GS3_SIZE_Y	480
#define	GS4_SIZE_Y	512

#define	GS1_MAX_COLOR	16
#define	GS2_MAX_COLOR	256
#define	GS3_MAX_COLOR	256
#define	GS4_MAX_COLOR	256

#define	GS1_MAX_CELL	128
#define	GS2_MAX_CELL	192
#define	GS3_MAX_CELL	256
#define	GS4_MAX_CELL	256

#define SIZE_X_218	448
#define SIZE_Y_218	320

#define SIZE_X_224C	640
#define SIZE_Y_224C	400

#define FILE_MARK_CELL	0x20
#define FILE_MARK_PALET	0x10


/* xkiss parameters */

#define DEFAULT_SIKII_DITHER	36	/* if bigger, turn black */
#define DEFAULT_SIKII_GOSA	1300
#define DEFAULT_COEF_GOSA	25

#define	MAX_CELL	GS4_MAX_CELL
#define	MAX_SET		10
#define	MAX_PAL		10
#define	MAX_COLOR	GS4_MAX_COLOR

#define MAX_SIZE_X      GS4_SIZE_X
#define MAX_SIZE_Y	GS4_SIZE_Y
#define DEFAULT_SIZE_X  SIZE_X_218
#define DEFAULT_SIZE_Y	SIZE_Y_218

#define	FILENAME_LENGTH	128


/* button drawing parameters */

#define	BTN_WIDTH	17
#define	BTN_HEIGHT	16
#define QUIT_WIDTH	40
#define	BTN_SPACE	2
#define	GROUP_SPACE	10
#define	FONT_X		5
#define	FONT_Y		14

#define BTN_Y_OFFSET (BTN_HEIGHT+BTN_SPACE*2+1)
#define BTN_X_SET    (BTN_SPACE)
#define BTN_X_PAL    (BTN_SPACE+GROUP_SPACE+(BTN_WIDTH+BTN_SPACE)*MAX_SET)
#define BTN_X_QUIT   (BTN_SPACE+GROUP_SPACE*2 + \
		      (BTN_WIDTH+BTN_SPACE)*(MAX_SET+MAX_PAL))

#define FONTNAME	"-*-helvetica-bold-r-normal--14-*-iso8859-1"

#define DEFAULT_BORDER_WIDTH	2


/* valiables declaration */

typedef enum {OLD,NEW} FORMAT;

typedef struct {
    char		name[FILENAME_LENGTH];
    FORMAT		format;
    int			bit_per_pixel;
    int			xsize,ysize;
    int			xoffset,yoffset;
    int			mark;
    int			init_fix,fix;
    int			palet_num;
    enum {NOUSE,USE}	set[MAX_SET];
    unsigned char 	*buf,*mono_buf;
    int			x[MAX_SET],y[MAX_SET];
    Pixmap		pix,clip_pix;
    GC			gc,*paint_gc[MAX_PAL],gc_bw[2];
    int			pal;
} CELL;

CELL cell[MAX_CELL];


typedef struct {
    int red,green,blue;
    int pixel,mono;
} COLOR;

typedef struct {
    char	name[FILENAME_LENGTH];
    FORMAT	format;
    int		color_num;
    int		palet_num;
    int		bit_per_pixel;
    int		pb[MAX_PAL];
    COLOR	*color[MAX_PAL];
} PALET;

PALET *palet[MAX_COLOR];

int	pb[MAX_SET];
int	outor_palet; /* set but no use */
int	palet_num;

int	current_set=0;

char	*disp_name=NULL;
char	*geometry_string=NULL;
Display	*disp;
int	scr;
Window	w,rootwin;
Pixmap	world_pix,mono_background_pix;
GC	world_gc,mono_background_gc;
unsigned long	w_pixel,b_pixel;

int	init_fix[MAX_CELL],fix[MAX_CELL];

int	cnf_size_x=DEFAULT_SIZE_X, cnf_size_y=DEFAULT_SIZE_Y;
int	size_x    =DEFAULT_SIZE_X, size_y    =DEFAULT_SIZE_Y;

int	sikii_dither = DEFAULT_SIKII_DITHER; 
int	sikii_gosa   = DEFAULT_SIKII_GOSA;
int	coef_gosa    = DEFAULT_COEF_GOSA;

#define SYS_COL	3
GC	gc_button[SYS_COL];
typedef	enum	BUTTON_COLOR {BLACK=0,GREEN=1,ORANGE=2} SYS_COLOR;
char	*color_name[]= {"black","palegreen","tan2"};
char	*mono_color_name[] = {"black","white","black"};
char	*mono_reverse_color_name[] = {"white","black","white"};

enum GROUP { SET,PAL,QUIT };
STATE	quit_state=OFF;

GC	root_box_gc;
int	root_sx,root_sy,root_ex,root_ey;
STATE	root_flag;
STATE	root_paste_flag=OFF;

int	press_mark=-1,press_cell;
int	press_dx,press_dy;
int	fix_press_x,fix_press_y;
STATE	pesi_flag;



char	path[256],config_file[100];

/*
 * Utility
 */

void check_str(char *str)
{
    char *p;
    for (p=str;*p!='\0';p++) {
	if (*p=='\n' || *p=='\r' || *p=='\t')	*p=' ';
	if (*p==';') {
	    *p='\0';
	    break;
	}
    }
}

char *skip_str(char *str)
{
    char *p;
    for (p=str;*p==' ';p++);
    for (;*p!=' ';p++)
	if (*p=='\0') return(p);
    for (;*p==' ';p++);
    return(p);
}

void usage(char *fn)
{
    if (fn!=NULL) fprintf(stderr,"Can't find file (%s).\n",fn);
    fprintf(stderr,"xkiss 2.00 test c (GS4)\n");
    fprintf(stderr,"Usage : xkiss [option][load_path][config_file]\n");
    fprintf(stderr,"[option]:\n");
    fprintf(stderr,"\t-b               :big mode\n");
    fprintf(stderr,"\t-i               :iconic mode\n");
    fprintf(stderr,"\t-d display       :set display name\n");
    fprintf(stderr,"\t-g geom          :set geometry string\n");
    fprintf(stderr,"\t-c               :color mode\n");
    fprintf(stderr,"\t-m num           :monochrome mode(default %d)\n",
	    DEFAULT_SIKII_DITHER);
    fprintf(stderr,"\t-e [num1[,num2]] :monochrome mode(default %d,%d)\n",
	    DEFAULT_SIKII_GOSA,DEFAULT_COEF_GOSA);
    fprintf(stderr,"\t-r               :reverse mode(only monochrome mode)\n");
    fprintf(stderr,"\t-q               :quiet mode\n");
    exit(-1);
}

FILE *msdos_fopen(char *fn,char *mode)
{
    FILE	*fp;
    char	*p;
    char	pathname[256],filename[256],*f;

    sprintf(filename,"%s",fn);
    check_str(filename);
    for (p=filename;*p!=' ' && *p!='\0';p++);
    *p='\0';
    
    sprintf(pathname,"%s%s",path,filename);
    if ((fp=fopen(pathname,mode))!=NULL)
	return(fp);
    
    for (p=filename;*p!='\0';p++)
	if (islower(*p)) *p=toupper(*p);
    sprintf(pathname,"%s%s",path,filename);
    if ((fp=fopen(pathname,mode))!=NULL)
	return(fp);
    
    for (p=filename;*p!='\0';p++)
	if (isupper(*p)) *p=tolower(*p);
    sprintf(pathname,"%s%s",path,filename);
    if ((fp=fopen(pathname,mode))!=NULL)
	return(fp);

    return(NULL);
}

FILE *kiss_fopen(char *fn,char *mode)
{
    FILE	*fp;

    if ((fp=msdos_fopen(fn,mode))==NULL) {
	usage(fn);
	exit(0);
    }
    return(fp);
}

unsigned char *kiss_malloc(int size)
{
    unsigned char *ptr;

    ptr=(unsigned char *)malloc(size);
    if (ptr==NULL) {
	printf("can't malloc. size %d\n",size);
	exit(0);
    }
    return(ptr);
}

/*
 * Data load
 */

int check_kiss_header(FILE *fp)
{
    unsigned char buf[4];

    fread(buf,1,4,fp);
    rewind(fp);
    if (buf[0]=='K' && buf[1]=='i' && buf[2]=='S' && buf[3]=='S')
	return(1);
    return(0);
}

void load_data_cell(char *fn,int num)
{
    FILE	*fp;
    unsigned char	buf[32],*tmp_buf;
    int		bit_per_pixel;
    int		i,j;
    int		xs,ys;
    int		sx,sy,ex,ey,x,y,flag;

    fp=kiss_fopen(fn,"r");
    if (check_kiss_header(fp)==1) {
	fread(buf,1,32,fp);
	if (buf[4]!=FILE_MARK_CELL) {
	    fprintf(stderr,"(%s) is not CELL file!\n",fn);
	    exit(0);
	}
	xs=buf[ 9]*256+buf[ 8];
	ys=buf[11]*256+buf[10];
	cell[num].xoffset=buf[13]*256+buf[12];
	cell[num].yoffset=buf[15]*256+buf[14];
	bit_per_pixel=buf[5];
	cell[num].format=NEW;
    } else {
	fread(buf,1,4,fp);
	xs=buf[1]*256+buf[0];
	ys=buf[3]*256+buf[2];
	cell[num].xoffset=0;
	cell[num].yoffset=0;
	bit_per_pixel=4;
	cell[num].format=OLD;
    }

    tmp_buf=kiss_malloc(xs*ys+1);
    for (j=0;j<ys;j++) {
	unsigned char *p,b[MAX_SIZE_X];
	switch (bit_per_pixel) {
	case 4:
	    fread(b,1,(xs+1)/2,fp);
	    p=tmp_buf+xs*j;
	    for (i=0;i<(xs+1)/2;i++) {
		*p++=(b[i]>>4)&0x0f;
		*p++=b[i]&0x0f;
	    }
	    break;
	case 8:
	    fread(tmp_buf+xs*j,1,xs,fp);
	    break;
	default:
	    fprintf(stderr,"%s : unknown format. bit/pixel = %d\n",
		    fn,bit_per_pixel);
	    exit(0);
	}
    }

    for (flag=0,sy=0;sy<ys;sy++) {
	for (x=0;x<xs;x++) 
	    if (*(tmp_buf+xs*sy+x)) {
		flag=1;
		break;
	    }
	if (flag==1) break;
    } 

    for (flag=0,ey=ys-1;ey>=0;ey--) {
	for (x=0;x<xs;x++) 	
	    if (*(tmp_buf+xs*ey+x)) {
		flag=1;
		break;
	    }
	if (flag==1) break;
    } 

    for (flag=0,sx=0;sx<xs;sx++) {
	for (y=0;y<ys;y++) 
	    if (*(tmp_buf+xs*y+x)) {
		flag=1;
		break;
	    }
	if (flag==1) break;
    } 

    for (flag=0,ex=xs-1;ex>=0;ex--) {
	for (y=0;y<ys;y++) 
	    if (*(tmp_buf+xs*y+x)) {
		flag=1;
		break;
	    }
	if (flag==1) break;
    } 

    if (big_mode==OFF)	cell[num].buf = kiss_malloc((ex-sx+1)*(ey-sy+1)  );
    else		cell[num].buf = kiss_malloc((ex-sx+1)*(ey-sy+1)*4);

    if (big_mode==OFF) {
	unsigned char *p,*t;
	p=cell[num].buf;
	for (j=sy;j<=ey;j++) {
	    t=tmp_buf+j*xs+sx;
	    for (i=sx;i<=ex;i++) {
		*p++=*t++;
	    }
	}
    } else {
	unsigned char *p1,*p2,*t;
	for (j=0;j<=ey-sy;j++) {
	    p1=&cell[num].buf[(ex-sx+1)*2*(j*2)];
	    p2=&cell[num].buf[(ex-sx+1)*2*(j*2+1)];
	    t=tmp_buf+(sy+j)*xs+sx;
	    for (i=sx;i<=ex;i++) {
		*p1++=*t;
		*p1++=*t;
		*p2++=*t;
		*p2++=*t;
		t++;
	    }
	}
    }
    free(tmp_buf);
    fclose(fp);
    cell[num].xsize = ex-sx+1;
    cell[num].ysize = ey-sy+1;
    cell[num].xoffset += sx;
    cell[num].yoffset += sy;
    if (big_mode==ON) {
	cell[num].xsize *=2;	
	cell[num].ysize *=2;
	cell[num].xoffset *=2;
	cell[num].yoffset *=2;
    } 
    cell[num].bit_per_pixel = bit_per_pixel;
}


void load_palet(char *fn,int n)
{
    FILE	*fp;
    unsigned char	buf[32];
    int		p,c;
    int		r,g,b;

    check_str(fn);
    fp=kiss_fopen(fn,"r");
    palet[n]=(PALET *)kiss_malloc(sizeof(PALET));
    sscanf(fn,"%s",palet[n]->name);
    if (check_kiss_header(fp)) {
	/* new format */
	fread(buf,1,32,fp);
	if (buf[4]!=FILE_MARK_PALET) {
	    fprintf(stderr,"(%s) is not PALET file!\n",fn);
	    exit(0);
	}
	palet[n]->color_num=buf[ 9]*256+buf[ 8];
	palet[n]->palet_num=buf[11]*256+buf[10];
	palet[n]->bit_per_pixel=buf[5];
	palet[n]->format=NEW;
    } else {
	/* old format */
	palet[n]->color_num=GS1_MAX_COLOR;
	palet[n]->palet_num=MAX_PAL;
	palet[n]->bit_per_pixel=12;
	palet[n]->format=OLD;
    }
    for (p=0;p<palet[n]->palet_num;p++) {
	palet[n]->color[p] = 
	    (COLOR *)kiss_malloc(sizeof(COLOR)*palet[n]->color_num);
	for (c=0;c<palet[n]->color_num;c++) {
	    switch (palet[n]->bit_per_pixel) {
		unsigned char	byte_h,byte_l,col;
	    case 12:
		fread(&byte_l,1,1,fp);
		fread(&byte_h,1,1,fp);
		r = ((byte_l>>4)&0x0f) * 0x11;
		g = ((byte_h   )&0x0f) * 0x11;
		b = ((byte_l   )&0x0f) * 0x11;
		break;
	    case 24:
		fread(&col,1,1,fp);
		r=(int)col;
		fread(&col,1,1,fp);
		g=(int)col;
		fread(&col,1,1,fp);
		b=(int)col;
		break;
	    default:
		fprintf(stderr,"%s : unknown format. bit/pixel = %d\n",
			fn,palet[n]->bit_per_pixel);
		exit(0);
	    }

	    if (color_mode==ON) {
		palet[n]->color[p][c].red   = r*0x0101;
		palet[n]->color[p][c].green = g*0x0101;
		palet[n]->color[p][c].blue  = b*0x0101;
	    }
	    if (mono_dither_mode==ON) {
		palet[n]->color[p][c].mono=((g*59 + r*30 + b*11)/33)/0x11;
	    }
	    if (mono_gosa_mode==ON) {
		double	m;
		int	mono_palet;
		m=1.0*(g*59 + r*30 + b*11)/0x11/sikii_gosa;
		mono_palet = ((m*m-1.0)*coef_gosa/100 + m)*sikii_gosa;
		if (mono_palet<0)		mono_palet=0;
		if (mono_palet>sikii_gosa)	mono_palet=sikii_gosa;
		palet[n]->color[p][c].mono=mono_palet;
	    }
	}
    }
    fclose(fp);
    for (p=0;p<MAX_PAL;p++)
	palet[n]->pb[p]=0;
    for (p=0;p<palet[n]->palet_num;p++)
	palet[n]->pb[p]=p;
}

void load_config_set(int set,char *str)
{
    int		n;
    char	*p=str;
    
    check_str(str);
    sscanf(str,"%d",&pb[set]);
    for (n=0;n<MAX_CELL;n++) {
	p=skip_str(p);
	if (*p!='*') {
	    int	i,x,y;
	    if (sscanf(p,"%d,%d",&x,&y)!=2) break;
	    if (big_mode==ON) {
		x*=2;
		y*=2;
	    }
	    for (i=0;i<MAX_CELL;i++)
		if (cell[i].mark==n) {
		    cell[i].x[set]=x;
		    cell[i].y[set]=y;
		}	
	}
    }	
}

void load_config_size(char *str)
{
    char	*p;
    for (p=str;*p!='\0';p++)
	if (!isdigit(*p)) *p=' ';
    sscanf(str,"%d %d",&cnf_size_x ,&cnf_size_y);
    size_x=cnf_size_x;
    size_y=cnf_size_y;
}

void load_config_cell(int num,char *str)
{
    char	*p,*pp,c;
    int		i,j,n;
    
    check_str(str);
    sscanf(str,"%d",&cell[num].mark);

    for (p=str;;p++)
	if (!isdigit(*p)) break;
    if (*p=='.') {
	p++;
	for (pp=p;;p++)
	    if (*p==' ' || (!isdigit(*p)))	
		break;
	c=*p;
	*p='0';
	sscanf(pp,"%d",&j);
	cell[num].init_fix=j;
	cell[num].fix=j;
	*p=c;
    }
    for (;;p++)
	if (*p!=' ') break;

    sscanf(p,"%s",cell[num].name);
    str=skip_str(p);
    
    load_data_cell(cell[num].name,num);
    
    cell[num].palet_num=0;
    for (p=str;*p!='\0';p++)
	if (*p=='*') {
	    sscanf(p+1,"%d",&cell[num].palet_num);
	    break;
	}

    for (p=str;*p!='\0';p++)
	if (*p==':') {
	    p++;
	    for (i=0;i<MAX_SET;i++)
		cell[num].set[i]=NOUSE;
	    for (;;) {
		if (sscanf(p,"%d",&j)==0) break;
		if (0<=j && j<MAX_SET)	cell[num].set[j]=USE;
		p=skip_str(p);
		if (*p=='\0') break;
	    }
	    return;
	}

    for (i=0;i<MAX_SET;i++)
	cell[num].set[i]=USE;
}

void arg_check(int argc,char **argv)
{
    char	*arg_str[20];
    int		arg=0,i;
    int		sikii,coef,sscanf_value;

    for (i=1;i<argc;i++) {
	if (argv[i][0]=='-') {
	    switch (argv[i][1]) {
	    case 'b': 
		big_mode     =(big_mode    ==ON) ? OFF : ON;	break;
	    case 'q':
		quiet_mode   =(quiet_mode  ==ON) ? OFF : ON;	break;
	    case 'i':
		iconic_mode  =(iconic_mode ==ON) ? OFF : ON;	break;
	    case 'r':
		reverse_mode =(reverse_mode==ON) ? OFF : ON;	break;
	    case 'd': 
		disp_name=argv[++i];		break;
	    case 'g':
		geometry_string=argv[++i];	break;
	    case 'c':
		color_mode       = ON;
		mono_dither_mode = OFF;
		mono_gosa_mode   = OFF;
		break;
	    case 'm':
		color_mode       = OFF;
		mono_dither_mode = ON;
		mono_gosa_mode   = OFF;
		if (sscanf(&argv[i][2],"%d",&sikii)==1) {
		    sikii_dither=sikii;
		} else {
		    if (i+1 < argc && sscanf(argv[i+1],"%d",&sikii)==1) {
			sikii_dither=sikii;
			i++;
		    }
		}
		break;
	    case 'e':
		color_mode       = OFF;
		mono_dither_mode = OFF;
		mono_gosa_mode   = ON;
		sscanf_value=sscanf(&argv[i][2],"%d,%d",&sikii,&coef);
		if (sscanf_value==2) {
		    sikii_gosa=sikii;
		    coef_gosa=coef;
		    break;
		}
		if (sscanf_value==1) {
		    sikii_gosa=sikii;
		    break;
		}
		if (i+1 == argc)
		    break; 
		sscanf_value=sscanf(&argv[i+1][0],"%d,%d",&sikii,&coef);
		if (sscanf_value==2) {
		    sikii_gosa=sikii;
		    coef_gosa=coef;
		    i++;
		    break;
		}
		if (sscanf_value==1) {
		    sikii_gosa=sikii;
		    i++;
		    break;
		}
		break;
	    default:
		usage(NULL);
	    }
	} else {
	    arg_str[arg++]=argv[i];
	}
    }

    if (arg==0) {
	sprintf(path,"./");
	sprintf(config_file,"kiss.cnf");
    }
    if (arg==1) {
	FILE	*fp;
	char	*p,*pp=NULL;
	sprintf(path,"%s/",arg_str[0]);
	sprintf(config_file,"kiss.cnf");
	if ((fp=msdos_fopen(config_file,"r"))!=NULL) {
	    fclose(fp);
	} else {
	    for (p=arg_str[0];*p!='\0';p++)	
		if (*p=='/') pp=p;
	    if (pp==NULL) {
		sprintf(path,"./");
		sprintf(config_file,"%s",arg_str[0]);
	    } else {
		*pp='\0';
		sprintf(path,"%s/",arg_str[0]);
		sprintf(config_file,"%s",pp+1);
	    }	    
	}
    }
    if (arg==2) {
	sprintf(path,"%s/",arg_str[0]);
	sprintf(config_file,"%s",arg_str[1]);
    }
    if (arg>=3) {
	fprintf(stderr,"argument error!\n");
	usage(NULL);
    }
}

void load_config()
{
    FILE	*fp;
    int		num=0,set=-1,i;
    char	str[1024];
    static char	set_str[MAX_SET][8*MAX_CELL+10];
    STATE	palet_load_flag;

    fp=kiss_fopen(config_file,"r");
    if (quiet_mode==OFF)
	fprintf(stderr,"loading config file (%s%s)...\n",path,config_file);

    for (i=0;i<MAX_CELL;i++) {
	cell[i].mark=-1;
	cell[i].init_fix=0;
	cell[i].fix=0;
	init_fix[i]=0;
	fix[i]=0;
    }

    palet_load_flag=OFF;
    palet_num=0;
    while (fgets(str,1024,fp)!=NULL) {
	switch (str[0]) {
	case '%':
	    load_palet(str+1,palet_num);
	    palet_num++;
	    palet_load_flag=ON;
	    break;
	case '[':
	    outor_palet=atoi(str+1);
	    break;
	case '(':
	    load_config_size(str+1);
	    break;
	case '#':
	    load_config_cell(num++,str+1);
	    break;
	case '$':
	    strcpy(set_str[++set],str+1);
	    break;
	case ' ':
	    strcat(set_str[set],str+1);
	    break;
	}
    }
    fclose(fp);

    if (palet_load_flag==OFF)	usage(config_file);

    if (big_mode==ON) {
	cnf_size_x*=2;
	cnf_size_y*=2;
	size_x*=2;
	size_y*=2;
    }
    for (i=0;i<=set;i++)
	load_config_set(i,set_str[i]);

    for (i=0;i<MAX_CELL;i++) {
	init_fix[i]=0;
	fix[i]=0;
    }
    for (i=0;i<MAX_CELL;i++) {
	int	f;
	if (cell[i].mark==-1) continue;
	f=cell[i].init_fix;
	if (init_fix[cell[i].mark]<f) {
	    init_fix[cell[i].mark]=+f;
	    fix[cell[i].mark]=+f;
	}
    }
}

void save()
{
    int		i,j,k,m,n=0;
    FILE	*fp_w,*fp_r;
    static char buf[2048];

    fp_w=fopen("xkiss.cnf","w");
    fp_r=kiss_fopen(config_file,"r");
   
    while (fgets(buf,2048,fp_r)!=NULL)
	if (buf[0]!='$' && buf[0]!=' ')
	    fprintf(fp_w,"%s",buf);
    fclose(fp_r);

    for (k=0;k<MAX_SET;k++) {
	fprintf(fp_w,"$%d ",pb[k]);
	for (i=0;i<MAX_CELL;i++) {
	    m=-1;
	    for (j=0;j<MAX_CELL;j++)
		if (cell[j].mark==i) m=j;
	    if (m==-1 || cell[m].set[k]==0) fprintf(fp_w,"* ");
	    else fprintf(fp_w,"%d,%d ",cell[m].x[k],cell[m].y[k]);
	    if (n++==16) {
		n=0;
		fprintf(fp_w,"\n ");
	    }
	}
	fprintf(fp_w,"\n");
    }
    fclose(fp_w);
}

/*
 * X Drawing Utility
 */

void quit()
{
    Visual	*vis;

/*  save(); */

    XDestroyWindow(disp,w);

    vis = DefaultVisual(disp,scr);
    if (root_paste_flag==ON) {
	if (vis->class !=StaticColor && vis->class!=StaticGray) {
	    Atom	atom;
	    Pixmap	pix;
	    atom = XInternAtom(disp, "_XSET_ROOT_ID", False);
	    pix = XCreatePixmap(disp, rootwin, 1,1,XDefaultDepth(disp,0));
	    XChangeProperty(disp, rootwin, atom, XA_PIXMAP, 32,
			    PropModeReplace,(unsigned char *)&pix, 1);
	    XSetCloseDownMode(disp, RetainTemporary);
	}
    }
    XCloseDisplay(disp);
    exit(0);
}

void monochrome_sub(unsigned char *cbuf,unsigned char *mbuf,
		    int dx,int dy,int p,int n)
{
    static int  d_tbl[4][4] = {
	{  0, 12,  3, 15 },
	{ 18,  6, 21,  9 },
	{  3, 15, 18, 12 },
	{ 21, 12, 18,  6 }
    };
    static int	tone_buf[2][3000];
    int		tone,carry,lb,lc;
    int		x,y;
    int		c,color;

    if (mono_dither_mode==ON) {
	for (y=0;y<dy;y++)
	    for (x=0;x<dx;x++) {
		c=(int)(*cbuf++);
		color=palet[n]->color[p][c].mono+d_tbl[y%4][x%4];
		*mbuf++ = (color>sikii_dither) ? 0:1;
	    }
    } else {
	for (y=0;y<2;y++)
	    for (x=0;x<dx;x++)
		tone_buf[y][x]=0;

	for (y=0;y<dy;y++) {
	    lb=(y+1)&1;
	    lc=y&1;
	    carry=0;
	    for (x=0;x<dx;x++)
		tone_buf[lb][x]=0;
	    for (x=0;x<dx;x++) {
		c=*cbuf++;
		tone = palet[n]->color[p][c].mono + tone_buf[lc][x] + carry;
		if (tone >= sikii_gosa) {
		    tone-=sikii_gosa;
		    *mbuf++=0;
		} else {
		    *mbuf++=1;
		}
		carry = tone >>3;
		tone_buf[lb][x] += carry + (tone & 7);
		tone_buf[lb][x>0 ? x-1 : x]+=carry<<1;
		tone_buf[lb][x<x-1 ? x+1:x]+=carry;
		carry<<=2;
	    }
	}
    }
}
void monochrome(int n,int p)
{
    monochrome_sub(cell[n].buf,cell[n].mono_buf,
		   cell[n].xsize,cell[n].ysize,p,cell[n].palet_num);
}

void make_mono_background(STATE flag)
{
    static int		mono_background_palet=-1;
    unsigned char	*cbuf,*mbuf,*b;
    int			x,y;

    if(palet[0]->color[pb[current_set]][0].mono==mono_background_palet &&
       flag==OFF) return;
    mono_background_palet=palet[0]->color[pb[current_set]][0].mono;

    cbuf=kiss_malloc(size_x*size_y*2);
    mbuf=kiss_malloc(size_x*size_y*2);
    b=cbuf;
    for (y=0;y<size_y*2;y++)
	for (x=0;x<size_x;x++)
	    *b++=0;
    monochrome_sub(cbuf,mbuf,size_x,size_y*2,pb[current_set],0);
    XSetForeground(disp,mono_background_gc,w_pixel);
    XFillRectangle(disp,mono_background_pix,mono_background_gc,
		   0,0,size_x,size_y);
    XSetForeground(disp,mono_background_gc,b_pixel);
    b=mbuf+size_x*size_y;
    for (y=0;y<size_y;y++)
	for (x=0;x<size_x;x++)
	    if ((*b++)==1)
		XDrawPoint(disp, mono_background_pix,mono_background_gc, x,y);
    free(mbuf);
    free(cbuf);
}

void make_palet()
{
    int         p,c,n,i,j;
    XColor	col;
    STATE	alloc_fail_flag=OFF;

    if (color_mode==OFF) {
	if (reverse_mode==OFF) {
	    w_pixel=WhitePixel(disp,scr);
	    b_pixel=BlackPixel(disp,scr);
	} else {
	    w_pixel=BlackPixel(disp,scr);
	    b_pixel=WhitePixel(disp,scr);
	}
	return;
    }

    for (n=0;n<palet_num;n++) {
	for (p=0;p<palet[n]->palet_num;p++) {
	    for (c=0;c<palet[n]->color_num;c++) {
		XColor xcolor;
		if (n!=0 && c==0) continue;
		xcolor.red  =palet[n]->color[p][c].red;
		xcolor.green=palet[n]->color[p][c].green;
		xcolor.blue =palet[n]->color[p][c].blue;
		if (XAllocColor(disp, DefaultColormap(disp,0),&xcolor)!=0) {
		    palet[n]->color[p][c].pixel=xcolor.pixel;
		} else {
		    palet[n]->color[p][c].pixel=-1;
		    alloc_fail_flag=ON;
		}
	    }
	}
    }

    if (alloc_fail_flag==ON) {
	XColor	*colors;
	int	color_max;
	color_max=1<<DefaultDepth(disp,scr);
	colors=(XColor *)kiss_malloc(sizeof(XColor)*color_max);
	for (i=0;i<color_max;i++)
	    colors[i].pixel=i;
	XQueryColors(disp,DefaultColormap(disp,0),colors,color_max);
	for (n=0;n<palet_num;n++) {
	    if (quiet_mode==OFF) {
		fprintf(stderr,"warning:color alloc failed (%s)\n",
			palet[n]->name);
	    }
	    for (p=0;p<palet[n]->palet_num;p++) {
		for (c=0;c<palet[n]->color_num;c++) {
		    double r,g,b,r2,g2,b2;
		    double e,e2;
		    int	cc=-1;
		    if (palet[n]->color[p][c].pixel!=-1) continue;
		    r2=(double)(palet[n]->color[p][c].red);
		    g2=(double)(palet[n]->color[p][c].green);
		    b2=(double)(palet[n]->color[p][c].blue);
		    e2=1.0e100;
		    for (i=0;i<color_max;i++) {
			r=(double)(colors[i].red);
			g=(double)(colors[i].green);
			b=(double)(colors[i].blue);
			e=(r-r2)*(r-r2)+(g-g2)*(g-g2)+(b-b2)*(b-b2);
			if (e<e2) {
			    e2=e;
			    cc=i;
			}
		    }
		    if (cc==-1) {
			if (quiet_mode==OFF) {
			    fprintf(stderr,
				    "color alloc failed. (%s) %d:%d\n",
				    palet[n]->name,p,c);
			}
		    } else {
			palet[n]->color[p][c].pixel=cc;
			if (quiet_mode==OFF) {
			    fprintf(stderr,
				    " %d %d",p,c);
			    fprintf(stderr,
				    "(%04x %04x %04x -> %04x %04x %04x) %d\n",
				    palet[n]->color[p][c].red,
				    palet[n]->color[p][c].green,
				    palet[n]->color[p][c].blue,
				    colors[cc].red,
				    colors[cc].green,
				    colors[cc].blue,
				    cc);
			}
		    }
		}
	    }
	}
	free(colors);
    }
}
void make_clip_pixmap()
{
    Pixmap	bpix;
    GC		bgc;
    int		xs,ys;
    int		x,y,n;

    unsigned char	*b;
    int		xp_num;
    XPoint	*xpts,*xpt;
  
#define	MAX_POINTS	2048	/* for XDrawPoints */

    xpts=(XPoint *)kiss_malloc(sizeof(XPoint)*MAX_POINTS);

    for (n=0;n<MAX_CELL;n++) {
	if (cell[n].mark==-1) continue;
	xs=cell[n].xsize;
	ys=cell[n].ysize;
	bpix=XCreatePixmap(disp,w,xs,ys,1);
	bgc=XCreateGC(disp,bpix,0,NULL);
	XSetForeground(disp,bgc,0);
	XFillRectangle(disp,bpix,bgc, 0,0, xs,ys);
	XSetForeground(disp,bgc,1);
	b=cell[n].buf;

	xpt=xpts;
	xp_num=0;
	for (y=0;y<ys;y++)
	    for (x=0;x<xs;x++) {
		if ((*b++)==0) continue;
		xpt->x=x;
		xpt->y=y;
		xpt++;
		xp_num++;
		if (xp_num==MAX_POINTS) {
		    XDrawPoints(disp,bpix,bgc,xpts,xp_num,CoordModeOrigin);
		    xpt=xpts;
		    xp_num=0;
		}
	    }
	if (xp_num>0)
	    XDrawPoints(disp,bpix,bgc,xpts,xp_num,CoordModeOrigin);
/*
        for (y=0;y<ys;y++)
            for (x=0;x<xs;x++)
                if ((*b++)!=0) XDrawPoint(disp,bpix,bgc,x,y);
*/
	cell[n].clip_pix=bpix;
	XFreeGC(disp, bgc);
    }
    free(xpts);
}

void repaint_pixmap(int new_pal,STATE init_mode)
{
    int		x,y,n,p;
    STATE	pal_array[MAX_COLOR];
    int		xs,ys;
    
    pal_array[0]=OFF;
    for (n=0;n<MAX_CELL;n++) {
	int	c,pn,cn,pal,npal;
	if (cell[n].mark==-1) continue;
	if (cell[n].set[current_set]==NOUSE) continue;
	if (cell[n].pal==new_pal) continue;
	pn=cell[n].palet_num;
	cn=palet[pn]->color_num;
	pal=cell[n].pal;
	npal=palet[pn]->pb[new_pal];
	for (c=1;c<cn;c++)
	    pal_array[c]=ON;
	if (cell[n].pal!=-1)
	    for (c=1;c<cn;c++) {
		if (color_mode==ON) {	/* color */
		    if (palet[pn]->color[npal][c].pixel
			==palet[pn]->color[pal][c].pixel)
			pal_array[c]=OFF;
		} else {		/* monochrome */
		    if (palet[pn]->color[npal][c].mono
			==palet[pn]->color[pal][c].mono)
			pal_array[c]=OFF;
		}
	    }
	xs=cell[n].xsize;
	ys=cell[n].ysize;
	if (color_mode==ON) {	/* color */
	    unsigned char *b;
	    int		c;
	    GC *col_gc=cell[n].paint_gc[npal];
	    b=cell[n].buf;
	    for (y=0;y<ys;y++)
		for (x=0;x<xs;x++) {
		    if (pal_array[c=*b++]==ON)
			XDrawPoint(disp, cell[n].pix,
				   cell[n].paint_gc[npal][c],x,y);
		}
	} else {		/* monochrome */
	    unsigned char	*b,*mb;
	    int		c;
	    monochrome(n,npal);
	    b=cell[n].buf;
	    mb=cell[n].mono_buf;
	    for (y=0;y<ys;y++)
		for (x=0;x<xs;x++) {
		    if (pal_array[c=*b++]==ON)
			XDrawPoint(disp,cell[n].pix,cell[n].gc_bw[*mb],x,y);
		    mb++;
		}
	}
	cell[n].pal=npal;
	if (init_mode==ON && quiet_mode==OFF) {
	    fprintf(stderr,".");
	    fflush(stderr);
	}
    }
}

void init_cell_pixmap()
{
    Pixmap	cpix;
    XGCValues	gcv;
    int		n,p,c,pn,x,y;
    int		pal,col;
    int		xs,ys;
    unsigned char	*b;

    make_clip_pixmap();

    for (n=0;n<MAX_CELL;n++) {
	if (cell[n].mark==-1) continue;
	xs=cell[n].xsize;
	ys=cell[n].ysize;
	cpix=XCreatePixmap(disp,w,xs,ys,XDefaultDepth(disp,0));
	cell[n].gc=XCreateGC(disp,cpix,0,NULL);
	if (color_mode==ON) {	/* color */
	    pn=cell[n].palet_num;
	    pal=palet[pn]->palet_num;
	    col=palet[pn]->color_num;
	    for (p=0;p<pal;p++) {
		cell[n].paint_gc[p]=(GC *)kiss_malloc(sizeof(GC)*col);
		for (c=0;c<col;c++) {
		    gcv.foreground=palet[pn]->color[p][c].pixel;
		    cell[n].paint_gc[p][c]=
			XCreateGC(disp,cpix,GCForeground,&gcv);
		}
	    }
	} else {		/* monochrome */
	    cell[n].gc_bw[0]=XCreateGC(disp,cpix,0,NULL);
	    cell[n].gc_bw[1]=XCreateGC(disp,cpix,0,NULL);
	    XSetForeground(disp,cell[n].gc_bw[0],w_pixel);
	    XSetForeground(disp,cell[n].gc_bw[1],b_pixel);
	    cell[n].mono_buf=kiss_malloc(xs*ys);
	}
	XSetClipMask(disp,cell[n].gc,cell[n].clip_pix);
	cell[n].pal=-1;
	cell[n].pix=cpix;
    }
    repaint_pixmap(pb[current_set],ON);
}

void set_back_color(STATE flag)
{
    if (color_mode==ON)
 	XSetForeground(disp,world_gc,
		       palet[0]->color[pb[current_set]][0].pixel);
    else
	make_mono_background(flag);
}

void remake_world_pixmap(int sx,int sy)
{
    static STATE	init_status=OFF;

    if (init_status==ON) {
	XFreeGC(disp,world_gc);
	XFreePixmap(disp,world_pix);
	if (color_mode!=ON) {
	    XFreeGC(disp,mono_background_gc);
	    XFreePixmap(disp,mono_background_pix);
	}
    }
    init_status=ON;
    size_x=sx;
    size_y=sy;
    world_pix=XCreatePixmap(disp,w, size_x,size_y, XDefaultDepth(disp,0));
    world_gc=XCreateGC(disp,world_pix,0,NULL);
    if (color_mode==OFF) {
	mono_background_pix=XCreatePixmap(disp,w,size_x,size_y,
					  XDefaultDepth(disp,0));    
	mono_background_gc=XCreateGC(disp,mono_background_pix,0,NULL);
    }
    set_back_color(ON);
}

void put_all()
{
    int	i;

    if (color_mode==ON)
	XFillRectangle(disp,world_pix,world_gc, 0,0, size_x,size_y);
    else 
	XCopyArea(disp,mono_background_pix,world_pix,mono_background_gc,
		  0,0, size_x,size_y, 0,0);
    for (i=MAX_CELL-1;i>=0;i--) {
	int	x,y;
	if (cell[i].set[current_set]==NOUSE) continue;
	x=cell[i].x[current_set]+cell[i].xoffset;
	y=cell[i].y[current_set]+cell[i].yoffset;
	XSetClipOrigin(disp,cell[i].gc,x,y);
	XCopyArea(disp,cell[i].pix,world_pix,cell[i].gc,
		  0,0, cell[i].xsize,cell[i].ysize, x,y);
    }    
    XCopyArea(disp,world_pix,w,world_gc, 0,0, size_x,size_y, 0,0);
}

void put_image(int pmark,int px,int py)
{
    int	left =size_x+1,right =-1,top =size_y+1,bottom =-1;
    int	left2=size_x+1,right2=-1,top2=size_y+1,bottom2=-1;
    int	sx,sy,ex,ey;
    int	i,p;

    for (i=MAX_CELL-1;i>=0;i--) {
	if (cell[i].set[current_set]==NOUSE) continue;
	if (cell[i].mark!=pmark) continue;
	sx=cell[i].x[current_set]+cell[i].xoffset;
	sy=cell[i].y[current_set]+cell[i].yoffset;
	ex=cell[i].xsize + sx;
	ey=cell[i].ysize + sy;
	if (sx     <left2) 	left2  =sx;
	if (right2 <ex   ) 	right2 =ex;
	if (sy     <top2 ) 	top2   =sy;
	if (bottom2<ey   )	bottom2=ey;
	ex=cell[i].xsize + px + cell[i].xoffset;
	ey=cell[i].ysize + py + cell[i].yoffset;
	if (px    <left) 	left  =px;
	if (right <ex  ) 	right =ex;
	if (py    <top ) 	top   =py;
	if (bottom<ey  ) 	bottom=ey;
	cell[i].x[current_set]=px;
	cell[i].y[current_set]=py;
    }
    if (left2 <left)	left  =left2;
    if (top2  <top)	top   =top2;
    if (right <right2)	right =right2;
    if (bottom<bottom2)	bottom=bottom2;
	
    if (left  <0     )	left  =0;
    if (right >size_x)	right =size_x;
    if (top   <0     )	top   =0;
    if (bottom>size_y)	bottom=size_y;

    if (color_mode==ON)
	XFillRectangle(disp,world_pix,world_gc, 
		       left2,top2, right2-left2,bottom2-top2);
    else 
	XCopyArea(disp,mono_background_pix,world_pix,mono_background_gc,
		  left2,top2, right2-left2,bottom2-top2,left2,top2);

    for (i=MAX_CELL-1;i>=0;i--) {
	int		cw,ch,cox,coy;
	if (cell[i].set[current_set]==NOUSE) continue;
	sx=cell[i].x[current_set]+cell[i].xoffset;
	sy=cell[i].y[current_set]+cell[i].yoffset;
	ex=cell[i].xsize + sx;
	ey=cell[i].ysize + sy;
	if (right<=sx ||bottom<=sy || ex<=left || ey<=top)	continue;
	cox= (sx<=left  ) ? left   : sx;
	cw = (ex<=right ) ? ex-cox : right-cox;
	coy= (sy<=top   ) ? top    : sy;
	ch = (ey<=bottom) ? ey-coy : bottom-coy;
	XSetClipOrigin(disp,cell[i].gc, sx,sy);
	XCopyArea(disp,cell[i].pix,world_pix,cell[i].gc,
		  cox-sx,coy-sy, cw,ch, cox,coy);
    }
    XCopyArea(disp,world_pix,w,world_gc, 
	      left,top, right-left,bottom-top, left,top);
}

void button_initialize() 
{
    XFontStruct	*font_info;
    char	**cname;
    int		i;

    font_info = XLoadQueryFont(disp,FONTNAME);
    if (color_mode==ON) {
	cname=color_name;
    } else {
	if (reverse_mode==OFF) cname=mono_color_name;
	else		   cname=mono_reverse_color_name;
    }
    for (i=0;i<SYS_COL;i++) {
	XColor	col,exact;
	if (XAllocNamedColor(disp,XDefaultColormap(disp,0),
			     cname[i],&exact,&col)==0) {
	    XAllocNamedColor(disp,XDefaultColormap(disp,0),
			     mono_color_name[i],&exact,&col);
	}
	gc_button[i] = XCreateGC(disp,w,0,NULL);
	XSetForeground(disp,gc_button[i],col.pixel);
	XSetFont(disp,gc_button[i],font_info->fid);
    }
}


void draw_button(enum GROUP group,int n,STATE state)
{
    SYS_COLOR 	color1,color2;
    int		x;
    static char	*number[10]={"0","1","2","3","4","5","6","7","8","9"};

    if (state==ON) {
	color1=BLACK;	color2=GREEN;
    } else {
	color1=GREEN;	color2=BLACK;
    }
    
    switch(group) {
    case SET:
    case PAL:
	x = ((group==SET) ? BTN_X_SET : BTN_X_PAL) + (BTN_WIDTH+BTN_SPACE)*n;
        
	XFillRectangle(disp,w,gc_button[color1],
		       x,size_y+BTN_SPACE,BTN_WIDTH,BTN_HEIGHT);
	XDrawRectangle(disp,w,gc_button[color2],
		       x,size_y+BTN_SPACE,BTN_WIDTH,BTN_HEIGHT);
	XDrawString(disp,w,gc_button[color2],
		    x+FONT_X,size_y+BTN_SPACE+FONT_Y,number[n],1);
	break;
    case QUIT:
        x = BTN_X_QUIT;
	XFillRectangle(disp,w,gc_button[color1],
		       x,size_y+BTN_SPACE,QUIT_WIDTH,BTN_HEIGHT);
	XDrawRectangle(disp,w,gc_button[color2],
		       x,size_y+BTN_SPACE,QUIT_WIDTH,BTN_HEIGHT);
	XDrawString(disp,w,gc_button[color2],
		    x+FONT_X,size_y+BTN_SPACE+FONT_Y,"Quit",4);
	break;
    }
}

void redisplay()
{
    int		i;
    
    XFillRectangle(disp,w,gc_button[ORANGE], 0,size_y, size_x,BTN_Y_OFFSET);

    for (i=0; i<MAX_SET; i++)
        draw_button(SET, i, (current_set==i) ? ON : OFF);
    for (i=0; i<MAX_PAL; i++)
        draw_button(PAL, i, (pb[current_set]==i) ? ON : OFF);
    draw_button(QUIT, 0, quit_state);
    
    put_all();
}

void check_command(int x,int y)
{
    int		i;
    
    if (y < size_y) return;

    if (x < BTN_X_SET) return;
    for (i=0;i<MAX_SET;i++)
	if (x<BTN_X_SET + (BTN_WIDTH+BTN_SPACE)*(i+1)) {
	    current_set=i;
	    repaint_pixmap(pb[current_set],OFF);
	    set_back_color(OFF);
	    redisplay();
	    return;
	}
    if (x < BTN_X_PAL) return;
    for (i=0;i<MAX_PAL;i++)
	if (x < BTN_X_PAL + (BTN_WIDTH+BTN_SPACE)*(i+1)) {
	    pb[current_set]=i;
	    repaint_pixmap(pb[current_set],OFF);
	    set_back_color(OFF);
	    redisplay();
	    return;
	}

    if (x < BTN_X_QUIT)	return;    
    quit_state=ON;
    redisplay();
    quit();
}

void button_press(int mx,int my)
{
    int		i;
    
    press_mark = -1;
    if (my>=size_y) return;
    for (i=0;i<MAX_CELL;i++) {
	int x,y,sx,sy;
	if (cell[i].mark==-1) continue;
	if (cell[i].set[current_set]==NOUSE) continue;
	x=cell[i].x[current_set];
	y=cell[i].y[current_set];
	sx=cell[i].xsize;
	sy=cell[i].ysize;
	if (x<=mx && mx<x+sx && y<=my && my<y+sy)
	    if (cell[i].buf[(my-y)*sx+(mx-x)]!=0) {
		press_cell=i;
		press_mark=cell[i].mark;
		break;
	    }
    }
    if (press_mark == -1) 
	return; /* no cell pressed */

    press_dx = mx-cell[press_cell].x[current_set];
    press_dy = my-cell[press_cell].y[current_set];
    fix_press_x=mx-press_dx;
    fix_press_y=my-press_dy;
    pesi_flag=OFF;
}

void button_motion(int mx,int my)
{
    int		num,i;
    
    if (press_mark==-1)	return;
    
    if (fix[press_mark]>0) {
	int	x=mx-fix_press_x-press_dx;
	int	y=my-fix_press_y-press_dy;
	if (pesi_flag==ON) return;
	if (((x>0) ? x : -x)+((y>0) ? y : -y)>=(big_mode==ON) ? 6:3) {
	    int	f;
	    f=init_fix[press_mark];
	    if (f>64)	f=64;
	    f=(f-fix[press_mark])/4;
	    if (f<0) f=0;
	    if (big_mode==ON) f*=2;
	    if (y<0 || y>f || x>f || x<-f) {
		x=y=0;
		pesi_flag=ON;
	    }
	    mx = x + fix_press_x + press_dx;
	    my = y + fix_press_y + press_dy;
	    put_image(press_mark,fix_press_x + x , fix_press_y + y);
	    return;
	}
    } else {
	if (mx<0)	mx=0;
	if (size_x<mx) 	mx=size_x;
	if (my<0) 		my=0;
	if (size_y<my) 	my=size_y;
	
	put_image(press_mark,mx-press_dx,my-press_dy);
    }
}

void button_release(int mx,int my)
{
    if (press_mark==-1)	return;

    if (fix[press_mark]>0) {
	int	x=mx-fix_press_x-press_dx;
	int	y=my-fix_press_y-press_dy;
	int	i;
	if (((x>0) ? x : -x)+((y>0) ? y : -y)>(big_mode==ON) ? 6:3) {
	    int	i,f;
	    f=fix[press_mark];
	    fix[press_mark]=0;
	    i=5;
	    do {
		f>>=1;
		fix[press_mark]+=f;
	    } while (--i);
	}
	if (x!=0 || y!=0) 
	    put_image(press_mark,fix_press_x,fix_press_y);
    }
}

void xor_rectangle(int sx,int sy,int ex,int ey)
{
    int		x,y,dx,dy;

    x = (sx<ex) ? sx : ex;
    y = (sy<ey) ? sy : ey;
    dx = (sx<ex) ? ex-sx : sx-ex;
    dy = (sy<ey) ? ey-sy : sy-ey;
    XDrawRectangle(disp,w,root_box_gc, x,y, dx,dy);
}

void root_press(int mx,int my)
{
    root_sx=root_ex=mx;
    root_sy=root_ey=my;
    root_flag=OFF;
}

void root_motion(int mx,int my)
{
    if (root_flag==ON)
	xor_rectangle(root_sx,root_sy,root_ex,root_ey);
    root_flag=ON;
    if (mx<0) mx=0;
    if (my<0) my=0;
    if (mx>size_x)	mx=size_x;
    if (my>size_y)	my=size_y;
    root_ex=mx;
    root_ey=my;
    xor_rectangle(root_sx,root_sy,root_ex,root_ey);
}

void root_release()
{
    Pixmap	bpix;
    GC		gc;
    int		sx,sy,dx,dy;

    xor_rectangle(root_sx,root_sy,root_ex,root_ey);
    sx = (root_sx<root_ex) ? root_sx : root_ex;
    sy = (root_sy<root_ey) ? root_sy : root_ey;
    dx = (root_sx<root_ex) ? root_ex-root_sx : root_sx-root_ex;
    dy = (root_sy<root_ey) ? root_ey-root_sy : root_sy-root_ey;
    if (dx==0 || dy==0) return;
    bpix=XCreatePixmap(disp,w, dx,dy, XDefaultDepth(disp,0));    
    gc=XCreateGC(disp,w,0,NULL);
    XCopyArea(disp,world_pix,bpix,gc, sx,sy, dx,dy, 0,0);
    XSetWindowBackgroundPixmap(disp,rootwin,bpix);
    XFreeGC(disp,gc);
    XFreePixmap(disp,bpix);
    XClearWindow(disp,rootwin);
    root_paste_flag=ON;
}

void main(int argc,char **argv)
{
    int		pos_x=-9999,pos_y=-9999;    
    int		i;
    
    arg_check(argc,argv);

    if ((disp = XOpenDisplay(disp_name))==NULL) {
	fprintf(stderr,"error! can't open display(%s).\n",disp_name);
	exit(0);
    }

    load_config();

    rootwin = DefaultRootWindow(disp);
    scr = DefaultScreen(disp);

    if (geometry_string!=NULL) {
	int	x,y,flag_mask;
	char	geometry_default[40];
	unsigned int width,height;
	sprintf(geometry_default,"%dx%d",cnf_size_x,cnf_size_y+BTN_Y_OFFSET);
	flag_mask=XGeometry(disp,scr, geometry_string, geometry_default,
			    DEFAULT_BORDER_WIDTH, 1,1, 0,0,
			    &x,&y,&width,&height);
	if ((flag_mask & XValue) || (flag_mask & XNegative)) pos_x = x;
	if ((flag_mask & YValue) || (flag_mask & YNegative)) pos_y = y;
	if (flag_mask & WidthValue)  cnf_size_x = (int)width;
	if (flag_mask & HeightValue) cnf_size_y = (int)height;
	if ((flag_mask & YNegative) && (flag_mask & HeightValue))
	    cnf_size_y -= BTN_Y_OFFSET;
    }
    w=XCreateSimpleWindow(disp,rootwin, pos_x,pos_y, 
			  cnf_size_x,cnf_size_y+BTN_Y_OFFSET,
			  DEFAULT_BORDER_WIDTH,0,0);

    root_box_gc=XCreateGC(disp,w,0,NULL);
    XSetFunction(disp,root_box_gc,GXxor);
    if (color_mode==ON)
	XSetForeground(disp,root_box_gc,XDisplayCells(disp,scr)-1);
    else 
	XSetForeground(disp,root_box_gc,1);

    make_palet();
    button_initialize();
    init_cell_pixmap();
    remake_world_pixmap(cnf_size_x,cnf_size_y);

    {
	XSizeHints	hints;
	hints.min_width  = size_x;
	hints.min_height = size_y+BTN_Y_OFFSET;
	hints.flags=PMinSize;
	if (pos_x!=-9999) {
	    hints.x = pos_x;
	    hints.y = pos_y;
	    hints.flags |= USPosition;
	}
	XSetNormalHints(disp,w,&hints);
    }
    if (iconic_mode==ON) {
	XWMHints	wmhint;
	wmhint.initial_state=IconicState;
	wmhint.flags=StateHint;
	XSetWMHints(disp,w,&wmhint);
    }

    XStoreName(disp,w,"xkiss");
    XSetIconName(disp,w,"xkiss");

    if (quiet_mode==OFF)
	fprintf(stderr, "\nLet's KISS !!\n");
    XSelectInput(disp,w,
		 ExposureMask | StructureNotifyMask | 
		 ButtonPressMask | ButtonMotionMask | ButtonReleaseMask);
    XMapWindow(disp,w);
    while(True) {
	XEvent	ev;
	static	int	button_status;

	XNextEvent(disp,&ev);
	switch(ev.type) {
	case Expose:
	    redisplay();
	    break;
	case ButtonPress:
	    switch (ev.xbutton.button) {
	    case Button2:
		root_press(ev.xbutton.x,ev.xbutton.y);
		break;
	    case Button3:
		/* if your X don't have XIconifyWindow(), deleate next line. */
		XIconifyWindow(disp,w,scr);
		break;
	    default:
		check_command(ev.xbutton.x,ev.xbutton.y);
		button_press(ev.xbutton.x,ev.xbutton.y);
		break;
	    }
	    button_status = ev.xbutton.button;
	    break;
	case MotionNotify:
	    switch (button_status) {
	    case Button2:
		root_motion(ev.xbutton.x,ev.xbutton.y);
		break;
	    default:
		button_motion(ev.xbutton.x,ev.xbutton.y);
		break;
	    }
	    break;
	case ButtonRelease:
	    switch (button_status) {
	    case Button1:
		button_release(ev.xbutton.x,ev.xbutton.y);
		break;
	    case Button2:
		root_release();
		break;
	    default:
		break;
	    }
	    break;
	case ConfigureNotify:
	    remake_world_pixmap(ev.xconfigure.width,
				ev.xconfigure.height-BTN_Y_OFFSET);
	    redisplay();
	    break;
	}
    }
}
